/** 
 *  @file   KBD430_DKS.c
 *  @brief  Driver for Digital Keyscan
 *  This layer handles the digital keyboard scanning, detecting key 
 *  press/release events, and reporting them to higher layers.
 */
/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/


//
// Include files
//
#include "KBD430_Keyboard_public.h"
#include "KBD430_hardware.h"
#include "KBD430_delaytimer.h"
#include "KBD430_HUT.h"

//
// Type definitions
//
/*! States of the Keyboard state machine */
typedef enum
{ 
    KBD_IDLE,           /*! Keyboard Idle (interrupt mode)  */
    KBD_FIRST_SCAN,     /*! Performing 1st scan             */
    KBD_WAIT_DEBOUNCE,  /*! Waiting between scans           */
    KBD_SECOND_SCAN,    /*! Performing 2nd scan             */
    KBD_PROCESS,        /*! Processing keys                 */
    KBD_WAIT_DELAY2,    /*! Waiting between scans           */
} KBD_StateM_t;

//
// Local variables
//
/*! KSO hardware mapping. Provides a Hardware Abstraction layer allowing
 *  for easier portability. Simply change the port and pin according to your
 *  hardware
 */
static const USBKBD_Port_t KSO_Pinmap[KSO_PINS] = {
    { KSO0_POUT,  KSO0_BIT},
    { KSO1_POUT,  KSO1_BIT},
    { KSO2_POUT,  KSO2_BIT},
    { KSO3_POUT,  KSO3_BIT},
    { KSO4_POUT,  KSO4_BIT},
    { KSO5_POUT,  KSO5_BIT},
    { KSO6_POUT,  KSO6_BIT},
    { KSO7_POUT,  KSO7_BIT},
    { KSO8_POUT,  KSO8_BIT},
    { KSO9_POUT,  KSO9_BIT},
    { KSO10_POUT, KSO10_BIT},
    { KSO11_POUT, KSO11_BIT},
    { KSO12_POUT, KSO12_BIT},
    { KSO13_POUT, KSO13_BIT},
    { KSO14_POUT, KSO14_BIT}
};

/*
 * Implemented keys table:
 * This array will be populated automatically based on USBKBD_scancodes_s/.
 * '1' represents a valid key (Bit7=Col7, Bit0 = Col0)
 * I.e.
 * 0xFF = All keys are valid for corresponding row
 * 0x03 = Key in Col1 and Col0 are valid for corresponding row
 * Note: this table is useful to detect potential ghost key conditions which
 *       can be discarded because some keys are not implemented
 */
uint8_t KBD_Validkeytable[KSO_PINS];

/*! External variable with scancode table */
extern const USBKBD_scancodest_t USBKBD_scancodes_s;
/*! Stores the value of each input port for each output pin */
__no_init static  hexksiarray_t    ksi_array[KSO_PINS];
/*! State machine for keyboard */
KBD_StateM_t KBD_StateM;

//
// Local prototypes
//
static bool KBD430_DKS_Scan(bool ReScan);
static uint8_t KBD430_DKS_Process(keyarray_t *key_array, uint8_t max_keys);
static bool isKeyMasked(hexksiarray_t *keyarray, uint8_t row, uint8_t col);


//
// Public functions
//
/******************************************************************************
*
 * @brief   Digital Keyscan initialization
 *  Initializes all registers, variables for digital keyscan
 *
 * @return  none
 *****************************************************************************/
void KBD430_DKS_Init(void)
{
    uint8_t i,j;
    
    // Set KSO as output to detect interrupts
    SET_KSO_INTERRUPT();

    // Check which keys are valid. 
    // Unimplemented keys can be ignored when checking for ghost keys 
    for (i=0; i < KSO_PINS; i++)
    {
        // All keys are valid by default
        KBD_Validkeytable[i] = 0xFF;

        for (j=0; j < KSI_PINS; j++)
        {
            // If a key is not implemented, mark it in GhostKeyTable
            if (USBKBD_scancodes_s.keycode[i*8+j] == hidUsageReserved)
            {
                KBD_Validkeytable[i] &= ~(1<<j);
            }
        }
    }
    
    // Check if a key is pressed
    if (KSI_READ() == 0x00)
    {
        // If no key pressed, wait for interrupt
        KSI_IES |= KSI_ALL;      // Set falling edge interrupt
        KSI_IFG &= ~KSI_ALL;      // clear flags
        KSI_IE |= KSI_ALL;        // Enable KSI interrupt
        KBD_StateM = KBD_IDLE;
    }
    else
    {
        // A key is pressed, work in polling mode
        SET_KSO_POLL();             // Set pins as inputs to start polling
        KSI_IE &= ~KSI_ALL;       // Disable interrupts
        KBD_StateM = KBD_FIRST_SCAN;
    }
}

/******************************************************************************
 *
 * @brief   Disable Digital Keyscan
 *  Disable the digital Keyscan
 *
 * @return  none
 *****************************************************************************/
void KBD430_DKS_Disable(void)
{
    // Disable DSO interrupt and timer
    SET_KSO_INTERRUPT();                // Set KSO as input with pull-down
    KSI_IE &= ~KSI_ALL;                 // Disable DKS interrupt
    KBD_StateM = KBD_IDLE;
}

/******************************************************************************
 *
 * @brief   Assert a KSO line (output low)
 *  
 * @param KSOLine  Selected KSO line 
 *
 * @return  none
 *****************************************************************************/
void KBD430_DKS_AssertKSO(uint8_t KSOLine)
{
    uint8_t * port_ptr;
    uint8_t * port_dir;

    // Set pointers for PxOUT and PxDIR registers
    port_ptr = (uint8_t *) KSO_Pinmap[KSOLine].pout;
    port_dir = (uint8_t *) KSO_Pinmap[KSOLine].pout + KSO_PDIR_OFFSET;
    // Set KSO pin as output Low
    *port_dir |= KSO_Pinmap[KSOLine].port_mask;
    *port_ptr &= ~KSO_Pinmap[KSOLine].port_mask;
    // Wait a few cycles before reading KSI 
    __delay_cycles(DELAY_DKS_KSO);
}


/******************************************************************************
 *
 * @brief   Deassert a KSO line (input)
 *  
 * @param KSOLine  Selected KSO line 
 *
 * @return  none
 *****************************************************************************/
void KBD430_DKS_DeAssertKSO(uint8_t KSOLine)
{
    uint8_t * port_dir;

    // Set pointers for PxOUT and PxDIR registers
    port_dir = (uint8_t *) KSO_Pinmap[KSOLine].pout + KSO_PDIR_OFFSET;
    // Set KSO pin as input
    *port_dir &= ~KSO_Pinmap[KSOLine].port_mask;
    // Wait a few cycles before reading KSI 
    __delay_cycles(DELAY_DKS_KSO);
}

/******************************************************************************
 *
 * @brief   Checks if a KSI line is pressed or not
 *  
 * @param KSOLine  Selected KSI line 
 *
 * @return  true if KSI is pressed, false if not
 *****************************************************************************/
bool KBD430_DKS_IsKSIPressed(uint8_t KSILine)
{
   
    // If KSI is not asserted for the corresponding bit, send false
    if (!(KSI_READ() & (1<<KSILine)))
    {
        return false;
    }
    else
    {
        return true;
    }
}

/******************************************************************************
 *
 * @brief   Handles the state machine when keyboard is in polling mode
 *
 * @return  true if scan is complete; false if not
 *****************************************************************************/
bool KBD430_DKS_PollingTask(keyarray_t * key_array, 
                               uint8_t *pressed_keys, 
                               uint8_t max_keys)
{
    bool ret = false; // Scan is not complete by default
    
    switch (KBD_StateM)
    {
        case KBD_FIRST_SCAN:
            KBD430_DKS_Scan(false);
            KBD_StateM = KBD_WAIT_DEBOUNCE;
            DelayTimer_DelayCycles_Start(DELAY_SCAN_CYCLES);   
        break;
        case KBD_WAIT_DEBOUNCE:
            if (DelayTimer_isComplete() == true)
            {
                DelayTimer_Clear();
                // When delay is complete, do second scan
                KBD_StateM = KBD_SECOND_SCAN;
            }
        break;
        case KBD_SECOND_SCAN:
            KBD430_DKS_Scan(true);
            KBD_StateM = KBD_PROCESS;
        break;
        case KBD_PROCESS:
            ret = true; // Scan is complete
            *pressed_keys = KBD430_DKS_Process(key_array, max_keys);
            if (*pressed_keys == 0)
            {
                // If no keys pressed, go back to interrupt mode
                KBD430_DKS_Init();
            }
            else
            {
                // If key pressed, repeat scan 
                DelayTimer_DelayCycles_Start(DELAY_SCAN_CYCLES);    
                KBD_StateM = KBD_WAIT_DELAY2;
            }
        break;
        case KBD_WAIT_DELAY2:
            if (DelayTimer_isComplete() == true)
            {
                DelayTimer_Clear();
                // When delay is complete, do new scan
                KBD_StateM = KBD_FIRST_SCAN;
            }
            
        break;
        case KBD_IDLE:
        default:
            // An interrupt will put keyboard in polling mode
        break;
    }
                
    return ret;
}


/******************************************************************************
 *
 * @brief   Digital Keyscan poll
 *  This function only reads the keys and they will be processed later by
 *  KBD430_DKS_Process().
 *
 * @param ReScan If false, perform first scan
 *               if true, re-check previous scan
 *
 * @return  true if any key is pressed, false if not
 *****************************************************************************/
static bool KBD430_DKS_Scan(bool ReScan)
{
    uint8_t row;
    uint8_t * port_ptr;
    uint8_t * port_dir;
    bool  key_pressed = false;

    __disable_interrupt();      // Disable interrupts during keyscan
    SET_KSO_POLL();             // Start with all KSO as input
    
    // Read all rows
    for (row=0; row < KSO_PINS; row++)
    {
        // Set pointers for PxOUT and PxDIR registers
        port_ptr = (uint8_t *) KSO_Pinmap[row].pout;
        port_dir = (uint8_t *) KSO_Pinmap[row].pout + KSO_PDIR_OFFSET;
        // Set KSO pin as output Low
        *port_dir |= KSO_Pinmap[row].port_mask;
        *port_ptr &= ~KSO_Pinmap[row].port_mask;
        // Wait a few cycles before reading KSI 
        __delay_cycles(DELAY_DKS_KSO);
        // Read the KSI pins (processed later to remove ghost keys)
        if (ReScan == false)
        {
            ksi_array[row] = KSI_READ();
        }
        else
        {
            ksi_array[row] &= KSI_READ();       // Check if key is pressed again
            ksi_array[row] &= KBD_Validkeytable[row]; // Remove unimplemented keys
            if (ksi_array[row] != 0x00)
            {
                key_pressed = true;   // A key (any) is pressed
            }
        }
        
        // Set KSO pin as input 
        *port_dir &= ~KSO_Pinmap[row].port_mask;
        // Set KSI pins as output high to bleed off the charge
        KSI_DIR |= KSI_ALL;
        KSI_OUT |= KSI_ALL;
        // Return KSI to Input
        KSI_DIR &= ~KSI_ALL;
    }
    
    SET_KSO_IDLE(); // Set KSO as input with pull-resistor to avoid floating pins
    
    __enable_interrupt();
    
    return key_pressed;
    
}


/******************************************************************************
 *
 * @brief   Processes the keys
 *  Processes the keys which were scanned by KBD430_DKS_Scan(), removes ghost
 *  keys and adds the key to an array which will be send to the report
 *
 * @param key_array  Pointer to an array where the detected keys will be stored
 *        max_keys   Maximum number of keys which can be stored in the array
 *
 * @return  number of detected keys (0 if no key is pressed)
 *****************************************************************************/
static uint8_t KBD430_DKS_Process(keyarray_t *key_array, uint8_t max_keys)
{
    uint8_t row, col, i;
    uint8_t key;
    uint8_t key_counter =0;

    // Process keys (row and col)
    for (row=0; row < KSO_PINS; row++)
    {
        // If a key is pressed in row, or was pressed before, check which column
        if (ksi_array[row] != 0x00) 
        {
             for (col =0; col < KSI_PINS; col++)
            {
                // Scan which kso is pressed in current ksi
                key = (row*KSI_PINS)+col;
                if (ksi_array[row] & (1<<col))
                {
                    // If col+row is pressed and ghost key
                    if (isKeyMasked(ksi_array,row,col) == false)
                    {
                        // A key was pressed and is not masked
                        if (key_counter < max_keys)
                        {
                            // Add key to array
                            key_array[key_counter++] = USBKBD_scancodes_s.keycode[key];
                        }
                    }
                    else
                    {
                        if (key_counter < max_keys)
                        {
                            // This is a ghost key
                            key_array[key_counter++] = hidUsageErrorRollOver;
                        }
                    }
                }
            }
        }
    }

    for (i=key_counter; i < max_keys; i++)
    {
        // Fill rest of the array with empty keys
        key_array[i] = 0x00;
    }
    
    return key_counter;
    
}

/******************************************************************************
 *
 * @brief   Checks for ghost keys
 *  checks if current row+col press is valid or it can be caused by potential
 *  ghost keys
 *
 *  @param keyarray  Pointer to the array of keys
 *  @param row       Row corresponding to key being checked
 *  @param col       Column corresponding to key being checked
 *
 * @return  true Key can be caused by a ghost key, false key is valid
 *****************************************************************************/
static bool isKeyMasked(hexksiarray_t *keyarray, uint8_t row, uint8_t col)
{
    bool ret = false;
    uint8_t i;
    uint16_t temp;

    for (i=0; i < KSO_PINS; i++)
    {
        temp = (keyarray[row] & keyarray[i]);   // Check keys in common
        
        // Check if there are other columns pressed
        if ( (i != row) &&
             ( ((temp & (1<<col)) != 0x00) &&
               ((keyarray[row] & keyarray[i]) != (1<<col)) ) )
        {
            // Report if there's a potential ghost key (current col and row
            // is already pressed) so that the key is ignored
            ret = true;
        }
    }

    return ret;
}

/******************************************************************************
 *
 * @brief   Checks if the DKS module can go to sleep 
 *  DKS can go to sleep when waiting for delay or when idle. The other states
 *  indicate that there's data to be processed 
 *
 * @return  true if module can sleep; false if it can't
 *****************************************************************************/
bool KBD430_DKS_CanSleep(void)
{
    if ((KBD_StateM == KBD_WAIT_DEBOUNCE) || 
        (KBD_StateM == KBD_WAIT_DELAY2) ||
          (KBD_StateM == KBD_IDLE))
    {
        return true;
    }
    else {
        return false;
    }
}

/******************************************************************************
*
 * @brief   KSI_Isr
 *  Wakes-up the mcu due to a key press and goes to polling mode
 *
 * @return  NONE
 *****************************************************************************/
#pragma vector=KSI_VECTOR
__interrupt void KSI_Isr(void)
{
    // Set KSO pins to start polling mode
    SET_KSO_POLL();
    KBD_StateM = KBD_FIRST_SCAN;
    KSI_IE &= ~KSI_ALL;

    LPM3_EXIT;     // exit LPM
    __no_operation();

}
