/** 
 *  @file   KBD430_Keyboard.c
 *  @brief  Driver for Keyboard
 *  This layer gets the keys from the DKS module and reports them to the 
 *  KBD430_Report module. It can handle special key combinations such as 
 *  Function (Fn) keys.
 */
/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/


//
// Include files
//
#include "KBD430_Keyboard_public.h"
#include "KBD430_HUT.h"
#include "KBD430_delaytimer.h"

//
// Macros
//
#define CONSUMER_KEY(x)     (0x1100 | x)    /*! Report a Consumer Control Key */
#define WRC_KEY(x)          (0x2200 | x)    /*! Report a Wireless Radio key   */

//
// Local vazriables
// 
/*! Stores the current pressed keys */
__no_init static  keyarray_t        key_array[MAX_KEYS];
/*! Stores the previosly pressed keys */
__no_init static  keyarray_t        key_array_prev[MAX_KEYS];
/*! Number of pressed keys */
__no_init static uint8_t            keysPressed;
/*! Shows if Function key is pressed */
__no_init static bool fnKeyPressed;

//
// Local prototypes
//
static bool CheckforGhostKeys(void);
static bool CheckforFnKey(void);
static void UpdateHIDReport(void);
static bool FlushHIDReport(void);
static uint16_t GetFnKey(key_t keyvalue);

//
// Public functions
//
/******************************************************************************
*
 * @brief   Keyboard initialization
 *  Initializes Digital keyscan and report.
 *  Must be called by the application before any other functions.
 *
 * @return  none
 *****************************************************************************/
void KBD430_Keyboard_Init(void)
{
    uint16_t i;
    KBD430_DKS_Init();       // Initialize keyscan
    KBD430_Report_Init();    // Initialize USB Keyboard report]
    
    for (i=0; i < MAX_KEYS; i++)
    {
        key_array[i] = 0x00;
        key_array_prev[i] = 0x00;
    }
    
}

/******************************************************************************
 *
 * @brief   Runs the state machine of the keyboard
 *  Must be called by the application when idle, or after waking-up
 *
 * @return  none
 *****************************************************************************/
void KBD430_Keyboard_Task(void)
{
    // Run the state machine of Digital Keyscan 
    if (KBD430_DKS_PollingTask(key_array, &keysPressed, MAX_KEYS) == true)
    {
        if (keysPressed == 0)
        {
            // If no keys are pressed, make sure the report is cleared
            if (FlushHIDReport() == true)
            {
                UpdateHIDReport();
            }
        }
        else
        {
      
            // If the keyscan is complete, update the reports
            // First check if key is a Function key
            fnKeyPressed = CheckforFnKey();
      
            // Check if there are ghost keys
            CheckforGhostKeys();
        
            // Update HID Report with new pressed keys or released keys
            UpdateHIDReport();
        }
        
    }
}


/******************************************************************************
 *
 * @brief   Flushes the HID report, cleaning all keys
 *
 * @return  true if any key was removed, false if report was already empty
 *****************************************************************************/
static bool FlushHIDReport(void)
{
    uint16_t i;
    // By default return false to indicate that report is empty
    bool ret = false;   
    
    for (i=0; i < MAX_KEYS; i++)
    {
        // Clear the buffer
        key_array[i] = 0x00;
        // Check if any previous keys were pressed
        if (key_array_prev[i] != 0x00)
        {
            ret = true; // Return true if a key was removed
        }
    }
    
    return ret;
}


/******************************************************************************
 *
 * @brief   Checks if Function key is pressed
 *
 * @return  true if Fn is pressed, false if not
 *****************************************************************************/
static bool CheckforFnKey(void)
{
    uint16_t i;
    
    for (i=0; i < MAX_KEYS; i++)
    {
        // Check if Function key is pressed
        if (key_array[i] == hidUsageReservedFn)
        {
            // If there is a function key, set flag
            key_array[i] = 0x00;    // Remove key from array 
            return true;            // Return true if key found
        }
    }
    
    // Return false if Fn is not pressed
    return false;
}

/******************************************************************************
 *
 * @brief   Checks if any Ghost key is found
 *  This function invalidates the whole report when a ghost key is found, this
 *  is an optional feature
 *
 * @return  true if Ghost key found, false if not
 *****************************************************************************/
static bool CheckforGhostKeys(void)
{
    uint16_t i, j;
    
    for (i=0; i < MAX_KEYS; i++)
    {
        // Check if there are any Ghost keys
        if (key_array[i] == hidUsageErrorRollOver)
        {
            // If there is an invalid key (ghost key), invalidate report
            for (i=0; i < MAX_KEYS; i++)
            {
                key_array[j] = hidUsageErrorRollOver;
            }
            return true;    // At least one ghost key was found
        }
    }
    
    return false;       // No ghost key found
}


/******************************************************************************
 *
 * @brief   If at least 1 key changes, update the HID reports
 *
 * @return  none
 *****************************************************************************/
static void UpdateHIDReport(void)
{
    uint16_t i, j;
    uint16_t fnkey;
    bool key_match;
    
     // First check for released keys
    for (i=0; i < MAX_KEYS; i++)
    {
        if (key_array_prev[i] != 0x00)
        {
            key_match = false;
            for (j=0; j < MAX_KEYS; j++)
            {
                if (key_array[j] == key_array_prev[i])
                {
                    // Key is still in array
                    key_match = true;
                }
            }
            // Key is not in array, remove it
            if (key_match == false)
            {
                if (fnKeyPressed == true)
                {
                    fnkey = GetFnKey(key_array_prev[i]);
                    if ((fnkey & 0xFF00) == CONSUMER_KEY(0))
                    {   
#ifdef USE_CONSUMER_REPORT
                        // Remove the Consumer report Key
                        KBD430_ConsReport_Removekey((fnkey & 0x00FF));
#endif
                    }
                    else if ((fnkey & 0xFF00) == WRC_KEY(0))
                    {
#ifdef USE_WRC_REPORT
                        // Remove the Wireless report key
                        KBD430_WRCReport_Removekey((fnkey & 0x00FF));
#endif
                    }
                    else if (fnkey != hidUsageReserved)
                    {
                        // Remove regular key using Fn
                        KBD430_KBDReport_Removekey(fnkey);
                    }
                }
                else
                {
                    // Remove standard key
                    KBD430_KBDReport_Removekey(key_array_prev[i]);
                }
            }
        }
    }
    
    // Check for new keys
    for (i=0; i < MAX_KEYS; i++)
    {
        if (key_array[i] != 0x00)
        {
            key_match = false;
            for (j=0; j < MAX_KEYS; j++)
            {
                if (key_array[i] == key_array_prev[j])
                {
                    key_match = true;
                }
            }
            if (key_match == false)
            {
                if (fnKeyPressed == true)
                {
                    fnkey = GetFnKey(key_array[i]);
                    if ((fnkey & 0xFF00) == CONSUMER_KEY(0))
                    {
#ifdef USE_CONSUMER_REPORT
                        // Add key to Consumer Report
                        KBD430_ConsReport_Addkey((fnkey & 0x00FF));
#endif
                    }
                    else if ((fnkey & 0xFF00) == WRC_KEY(0))
                    {
#ifdef USE_WRC_REPORT
                        // Add key to Wireless Remote Control Report
                        KBD430_WRCReport_Addkey((fnkey & 0x00FF));
#endif
                    }
                    else if (fnkey != hidUsageReserved)
                    {
                        // Add standard key using Fn 
                        KBD430_KBDReport_Addkey(fnkey);
                    }

                }
                else
                {
                    // Add standard key
                    KBD430_KBDReport_Addkey(key_array[i]);
                }
            }
        }
    }
    
    // Backup array
    for (i=0; i < MAX_KEYS; i++)
    {
        key_array_prev[i] = key_array[i];
    }
}



/******************************************************************************
 *
 * @brief   Checks all valid key combination using the Function key
 *  Developers can add new Function key combinations in this switch statement.
 *  Note that the macros are only used to indicate the caller that the key
 *  combination must be treated differently (i.e. Consumer key)
 *
 * @param   keyvalue  Key pressed in combination with Fn
 *
 * @return  LSB: Key value. 
 *          MSB: If 0x00, standard key
 *               If CONSUMER_KEY, Consumer report (i.e. multimedia)
 *               If WRC_KEY, Wireless Remote Control 
 *****************************************************************************/
static uint16_t GetFnKey(key_t keyvalue)
{
    switch (keyvalue)
    {
        case hidUsageF11:
          return hidUsageKeypadNumlock;         // Fn + F11 = Num Lock
        case hidUsageF12:
          return hidUsageScrollLock;            // Fn + F12 = Scroll Lock
        case hidUsagePageUp:
          return hidUsageHome;                  // Fn + PgUp = Home
        case hidUsagePageDown:
          return hidUsageEnd;                   // Fn + PgDn = End
        case hidUsageF8:
          return CONSUMER_KEY(0x04);            // Fn + F8 = Audio Mute
        case hidUsageUpArrow:
          return CONSUMER_KEY(0x01);            // Fn + Up = Audio Volume increase
        case hidUsageDownArrow:
          return CONSUMER_KEY(0x02);            // Fn + Down = Audio Volume decrease
        case hidUsageLeftArrow:
          return CONSUMER_KEY(0x40);            // Fn + Left = Brightness Increase
        case hidUsageRightArrow:
          return CONSUMER_KEY(0x80);            // Fn + Right = Brighntess Decrease
        case hidUsageF3:
          return WRC_KEY(0x01);                 // Fn + F3 = Toggle Wireless
        default:
          return hidUsageReserved;
    }
}

/******************************************************************************
 *
 * @brief   Checks if the keyboard controller can sleep
 *  If nothing is pending, the controller can sleep
 *
 * @return  true if keyboard can sleep, false if not
 *****************************************************************************/
bool KBD430_Keyboard_CanSleep(void)
{
    if ((KBD430_DKS_CanSleep() == true) && 
        (KBD430_Report_CanSleep() == true) &&
        (DelayTimer_isComplete() == false))
    {
        return true;
    }
    else
    {
        return false;
    }
    
}
