/** 
 *  @file   KBD430_delaytimer.c
 *  @brief  Delay timer
 *          Used to wake-up the device periodically when running in polling mode
 */
/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
//
// Include files
//
#include "KBD430_public_defs.h"
#include "KBD430_hardware.h"
#include "KBD430_delaytimer.h"

//
// Variables
//
static DelayTimer_Status_t DT_Status;   /*! Status of the timer */

//
// Public functions
//
/******************************************************************************
 * @brief   Starts a delay timer waiting for the specified number of milliseconds
 *
 * @param ms Number of ms to wait
 *
 * @return  none
 *****************************************************************************/
void DelayTimer_DelayMs_Start( uint16_t ms )
{
    uint32_t cycles;
    // Clear flag
    DT_Status = DT_Inprogress;
    
    // Calculate number of cycles based on ACLK and milliseconds
    cycles = (ACLK_FREQ*ms)/1000;
    
    // Configure Timer Trigger TA0.0
    TA0CCR0 = cycles;
    TA0CCTL0 = CCIE;    // Enable Timer interrupt, clear flag
    TA0CTL = TASSEL_1 + MC_1 | TACLR;   // ACLK, started
}

/******************************************************************************
 * @brief   Starts a delay timer waiting for the specified number of cycles
 *   This is a simplified version of DelayTimer_DelayMs_Start simply because it 
 *   doesn't perform the multiplication/division
 *
 * @param cycles Number of cycles to wait
 *
 * @return  none
 *****************************************************************************/
void DelayTimer_DelayCycles_Start(uint16_t cycles)
{

    // Clear flag
    DT_Status = DT_Inprogress;
    
    // Configure Timer Trigger TA0.0
    TA0CCR0 = cycles;
    TA0CCTL0 = CCIE;    // Enable Timer interrupt, clear flag
    TA0CTL = TASSEL_1 + MC_1 | TACLR;   // ACLK, started
}


/******************************************************************************
 * @brief   Checks if Delay timer is complete
 *
 * @return  true if delay is complete, false if not
 *****************************************************************************/
bool DelayTimer_isComplete(void)
{
    if (DT_Status == DT_Complete)
    {
        return true;
    }
    else
    {
        return false;
    }
    
}


/******************************************************************************
*
 * @brief   Clear timer, setting it in Idle mode
 *
 * @return  NONE
 *****************************************************************************/
void DelayTimer_Clear(void)
{
    DT_Status = DT_Idle;      // Clear flag
}

/******************************************************************************
*
 * @brief   Delay Timer interrupt service routine
 *  Wakes-up the mcu, notifies that the delay is complete
 *
 * @return  NONE
 *****************************************************************************/
#pragma vector=TIMER0_A0_VECTOR
__interrupt void DelayTimerIsr(void)
{
    DT_Status = DT_Complete;      // Set flag
    TA0CTL &= ~MC_3;            // Stop timer
    LPM3_EXIT;                  // exit LPM
    __no_operation();

}
