/** 
 *  @file   KBD430_HIDKBD_report.c
 *  @brief  Builds the HID Keyboard report
 */
/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
//
// Include files
//
#include "KBD430_Keyboard_public.h"

#ifdef KBD_USE_USB
#include "USB_config/descriptors.h"
#include "USB_API/USB_Common/usb.h"        // USB-specific functions


#ifdef _HID_
    #include "USB_API/USB_HID_API/UsbHid.h"
#endif
#endif

//
// Local function prototypes
//
static bool FindEmptySlot( uint8_t * slot );


//
// Local variables
//
/*! Keyboard HID Report buffer */
__no_init static KBDReportBuffer_t KBDReport_buff;
#ifdef USE_CONSUMER_REPORT
/*! Consumer HID Report buffer */
__no_init static ConsReportBuffer_t ConsReport_buff;
#endif
#ifdef USE_WRC_REPORT
/*! WRC HID Report buffer */
__no_init static WRCReportBuffer_t WRCReport_buff;
#endif

//
// Public functions
//
/******************************************************************************
 * @brief   Initializes the HID Keyboard Buffer
 *  Initializes all queues and pointers
 *
 * @return  none
 *****************************************************************************/
void KBD430_Report_Init( void )
{
    uint8_t i;

    // Initialize all buffers and pointers
    KBDReport_buff.Update = false;
    KBDReport_buff.EmptySlots.byte = ALL_SLOTS;
    KBDReport_buff.Size = KBDREP_BUFF_SIZE;
    for (i=0; i <KBDREP_BUFF_SIZE; i++)
    {
        KBDReport_buff.Buff[i] = 0;
    }
    
#ifdef USE_CONSUMER_REPORT
    ConsReport_buff.Update = false;
    ConsReport_buff.Size = CONSREP_BUFF_SIZE;
    ConsReport_buff.Buff[CONSREP_ID_OFFSET] = CONSREP_ID;
    ConsReport_buff.Buff[CONSREP_DATA_OFFSET] = 0;
#endif
#ifdef USE_WRC_REPORT
    WRCReport_buff.Update = false;
    WRCReport_buff.Size = WRCREP_BUFF_SIZE;
    WRCReport_buff.Buff[WRCREP_ID_OFFSET] = WRCREP_ID;
    WRCReport_buff.Buff[WRCREP_DATA_OFFSET] = 0;
#endif
}



/******************************************************************************
 * @brief  Add key to the report buffer
 *   Checks the buffer to see if there's space and adds a new key
 *
 * @param keyvalue value of key being added to buffer
 *
 * @return  true if added, false if buffer full
 *****************************************************************************/
bool KBD430_KBDReport_Addkey( key_t keyvalue )
{
    uint8_t s;

    // Run this in case they key is already in buffer
    KBD430_KBDReport_Removekey(keyvalue);

    // Check if this is a modifier
    if ((keyvalue & 0xF0) == 0xE0)
    {
        // If so, add modifier to report
        KBDReport_buff.Buff[KBDREP_MOD_OFFSET] |= (1<<(keyvalue & 0x0F));
        KBDReport_buff.Update = true;
        return true;
    }
    else
    {
        // This is an element array
        if (FindEmptySlot(&s) == true)
        {
            // Write key in buffer
            KBDReport_buff.Buff[s] = keyvalue;
            // Clear emptyslot bit
            KBDReport_buff.EmptySlots.byte &= ~(1<<s);
            KBDReport_buff.Update = true;
            return true;
        }
        else
        {
            // Key wasn't added to report
            return false;
        }
    }

}

/******************************************************************************
 * @brief  Adds key to the consumer report
 *   Note that the consumer report is bit-based, so it sets/clears bits when
 *   a key is pressed
 *
 * @param key   Bits being set in the report (value will be ORed with the report)
 *
 * @return  true 
 *****************************************************************************/
#ifdef USE_CONSUMER_REPORT
bool KBD430_ConsReport_Addkey( uint8_t key )
{
    ConsReport_buff.Buff[CONSREP_DATA_OFFSET] |= key;
    ConsReport_buff.Update = true;
    return true;
}
#endif

/******************************************************************************
 * @brief  Adds key to the Wireless Remote Control report
 *   Note that the WRC report is bit-based, but the bit is toggled
 *
 * @param key   Bits being set in the report (value will be XORed with the report)
 *
 * @return  true 
 *****************************************************************************/

#ifdef USE_WRC_REPORT
bool KBD430_WRCReport_Addkey( uint8_t key )
{
    WRCReport_buff.Buff[WRCREP_DATA_OFFSET] ^= key;
    WRCReport_buff.Update = true;
    return true;
}
#endif

/******************************************************************************
 * @brief  Remove Key from buffer
 *  Checks if a key is in a buffer and removes it (sending a release to HID)
 *
 * @param keyvalue Key being removed from buffer
 *
 * @return  true if removed, false if not found
 *****************************************************************************/
bool KBD430_KBDReport_Removekey(key_t keyvalue)
{
    uint8_t i;

    // Check if this is a modifier
    if ((keyvalue & 0xF0) == 0xE0)
    {
        // remove modifier from report
        KBDReport_buff.Buff[KBDREP_MOD_OFFSET] &= ~(1<<(keyvalue & 0x0F));
        KBDReport_buff.Update = true;
    }
    for (i=KBDREP_DATA_OFFSET; i < KBDREP_BUFF_SIZE; i ++)
    {
        // Look for key in buffer
        if ( (((KBDReport_buff.EmptySlots.byte) & (1<<i)) == 0) &&
             (KBDReport_buff.Buff[i]) == keyvalue )
        {
            KBDReport_buff.EmptySlots.byte |= (1<<i);
            KBDReport_buff.Buff[i] = 0;
            KBDReport_buff.Update = true;
            return true;
        }
    }

    // Key wasn't found
    return false;
}

/******************************************************************************
 * @brief  Removes key from the consumer report
 *   Note that the consumer report is bit-based, so it sets/clears bits when
 *   a key is pressed
 *
 * @param key   Bits being set in the report (value will be ANDed with the report)
 *
 * @return  true 
 *****************************************************************************/
#ifdef USE_CONSUMER_REPORT
bool KBD430_ConsReport_Removekey( uint8_t key )
{
    ConsReport_buff.Buff[CONSREP_DATA_OFFSET] &= ~key;
    ConsReport_buff.Update = true;
    return true;
}
#endif

/******************************************************************************
 * @brief  Removes key from the Wireless Remote control report
 *   Note that the WRC report is bit-based, but the bit is toggled
 *
 * @param key   Bits being set in the report (value will be ANDed with the report)
 *
 * @return  true 
 *****************************************************************************/
#ifdef USE_WRC_REPORT
bool KBD430_WRCReport_Removekey( uint8_t key )
{
    // nothing to do since WRC only toggles wireless
    return true;
}
#endif

/******************************************************************************
 * @brief  Sends an updated report to HID
 *  Sends an updated report to HID interface if a key was pressed/released
 *
 * @return  None
 *****************************************************************************/
void KBD430_Report_Update(void)
{

    if (KBDReport_buff.Update == true)
    {
        if (KBD430_ReportSend((uint8_t *)&KBDReport_buff.Buff[0], HID_Keyboard) == true)
        {
            // Clear flag if the report was sent correctly
            KBDReport_buff.Update = false;
        }
    }
#ifdef USE_CONSUMER_REPORT    
    if (ConsReport_buff.Update == true)
    {
        if (KBD430_ReportSend((uint8_t *)&ConsReport_buff.Buff[0], HID_Consumer) == true)
        {
            // Clear flag if the report was sent correctly
            ConsReport_buff.Update = false;
        }
    }
#endif
#ifdef USE_WRC_REPORT
    if (WRCReport_buff.Update == true)
    {
        if (KBD430_ReportSend((uint8_t *)&WRCReport_buff.Buff[0], HID_WRC) == true)
        {
            // Clear flag if the report was sent correctly
            WRCReport_buff.Update = false;
        }
    }
#endif
    
}


/******************************************************************************
 * @brief   Checks if there are pending reports or if this module can sleep
 *  Checks the queues for pending data
 *
 * @return  bool false: Report pending to send, true: No reports to send
 *****************************************************************************/
bool KBD430_Report_CanSleep(void)
{
    if ( (KBDReport_buff.Update == false)  
#ifdef USE_CONSUMER_REPORT
        && (ConsReport_buff.Update == false) 
#endif
#ifdef USE_WRC_REPORT
        && (WRCReport_buff.Update == false) 
#endif
        )
    {
        return true;
    }
    else
    {
        return false;
    }
}


/******************************************************************************
 * @brief  Find empty slot in report buffer
 *
 * @param slot  Written with empty slot if found
 *
 * @return  true if slot found, false if buffer is full
 *****************************************************************************/
static bool FindEmptySlot( uint8_t * slot )
{
    uint8_t i;

    // Check if report buffer is full
    if (KBDReport_buff.EmptySlots.byte != 0x00)
    {
        // Find the empty slow
        for (i=KBDREP_DATA_OFFSET; i < KBDREP_BUFF_SIZE; i++)
        {
            if (((KBDReport_buff.EmptySlots.byte) & (1<<i)) != 0x00)
            {
                *slot = i;
                return true;
            }
        }
    }
    // Nothing found, return error
    return false;
}
