/** 
 *  @file   KBD430_main.c
 *  @brief  I2C Keyboard main file for G2xx4
 */
/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

//
// Include files
//
#include "KBD430_Keyboard_public.h"
#include "KBD430_hardware.h"

#include "keyboard_descriptors.h"
#include "hidi2c_types.h"
#include "hid_engine.h"
#include "hidi2c_settings.h"


//
// Local function prototypes
//
static void Init_Clock(void);
static void Init_Ports(void);


//
// Global variables
//
/*! Buffer used to send data to host */
uint8_t InputReport[INPUT_REPORT_SIZE];   
/*! Indicates if new data was received from hidi2c host */
volatile BOOL DataReceivedKBD;  
/*! Byte received from hidi2c interface */
uint8_t DataReceived;           
/*! indicates when device can go to sleep */
volatile uint8_t gotoSleep;   


//
// Public functions
//
/******************************************************************************
 * @brief   Main USB Keyboard function
 *  Initializes the MCU and then jumps to a main loop routine which:
 *    - Checks for pressed keys
 *    - Reports keyboard events to HID-keyboard interface
 *
 * @return  none
 *****************************************************************************/
void main( void )
{
    DataReceivedKBD = FALSE;
    // Stop watchdog timer to prevent time out reset
    WDTCTL = WDTPW + WDTHOLD;                           // Stop WDT

    Init_Clock();                                       // Init clocks
    Init_Ports();                                       // Init GPIOs

#ifdef ENABLE_SW_BSL
    // Check if BSL is requested
    // This routine checks if 'MSP' is pressed after reset. 
    // If so, it calls BSL mode
    SET_KSO_POLL();                         // Set KSO as idle
    __delay_cycles(100000);                 // Wait for pull-ups
    KBD430_DKS_AssertKSO(8);                // Assert KSO for 'M'
    if (KBD430_DKS_IsKSIPressed(7) == true) // Check if KSI for 'M' is pressed
    {
        KBD430_DKS_DeAssertKSO(8);          // Deassert KSO for 'M'
        KBD430_DKS_AssertKSO(14);            // Assert KSO for 'S'
        if (KBD430_DKS_IsKSIPressed(3) == true) // Check if KSI for 'S' is pressed
        {
            KBD430_DKS_DeAssertKSO(14);         // Deassert KSO for 'S'
            KBD430_DKS_AssertKSO(4);           // Assert KSO for 'P'
            if (KBD430_DKS_IsKSIPressed(1) == true) // Check if KSI for 'P' is pressed
            {
                // Toggle LEDs to signal entry to BSL mode
                LED_PORT_W &= ~(LEDNUM_2_1|LEDCAPS_2_2|LEDSCROLL_2_3);
                __delay_cycles(2000000);
                LED_PORT_W |= (LEDNUM_2_1|LEDCAPS_2_2|LEDSCROLL_2_3);
                __delay_cycles(2000000);
                LED_PORT_W &= ~(LEDNUM_2_1|LEDCAPS_2_2|LEDSCROLL_2_3);
                if ((FCTL3 & LOCKA) != 0x00)
                {
                    // BSL automatically toggles LOCKA before mass erase,
                    // By disabling protection here, BSL will re-enable it
                    // and it won't erase Segment A
                    FCTL3 = FWKEY | LOCKA;  
                }
                BCSCTL2 = 0x00;
                __bic_SR_register(0xFF);
                (*(void (**)())0x0C00)();         // Jump to BSL
            }
        }
    }
    SET_KSO_INTERRUPT();                        // Set KSO for interrupt
#endif    
    
    
    KBD430_Keyboard_Init();             // Initialize Digital Keyscan
    
    /* Creating our settings array */
    tHidI2CSetting settings[1];
    settings[0].bInterface = USCIB0;
    settings[0].pReportDescriptor = (uint8_t*) report_desc_HID0;
    settings[0].wReportLength = sizeof(report_desc_HID0);

    /* Initializing HID and I2C */
    if (initializeHidI2CStack(settings, 1))
    {
        // Error during initialization
        while(1)   
          ;
    }
    
    // Toggle LEDs just to signal that keyboard is ready
    LED_PORT_W &= ~(LEDNUM_2_1|LEDCAPS_2_2|LEDSCROLL_2_3);
    __delay_cycles(2000000);
    LED_PORT_W |= (LEDNUM_2_1);
    __delay_cycles(2000000);
    LED_PORT_W |= (LEDNUM_2_1|LEDCAPS_2_2);
    __delay_cycles(2000000);
    LED_PORT_W |= (LEDNUM_2_1|LEDCAPS_2_2|LEDSCROLL_2_3);
    __delay_cycles(2000000);
    LED_PORT_W &= ~(LEDNUM_2_1|LEDCAPS_2_2|LEDSCROLL_2_3);
    
    __enable_interrupt();   


    for (;;)
    {
        
        // This flag is true if the Standard keyboard interface gets data
        if (DataReceivedKBD == TRUE)
        {
            DataReceivedKBD = FALSE;
            // Check for LED usage data and turn on/off NUM, CAPS & SCROLL LEDs
            if (DataReceived & 0x01)
            {
                LED_PORT_W |= (LEDNUM_2_1);
            }
            else
            {
                LED_PORT_W &= ~(LEDNUM_2_1);
            }

            if (DataReceived & 0x02)
            {
                LED_PORT_W |= (LEDCAPS_2_2);
            }
            else
            {
                LED_PORT_W &= ~(LEDCAPS_2_2);
            }

            if (DataReceived & 0x04)
            {
                LED_PORT_W |= (LEDSCROLL_2_3);
            }
            else
            {
                LED_PORT_W &= ~(LEDSCROLL_2_3);
            }
        }
        
        KBD430_Keyboard_Task();      // Task checking keyboard
        KBD430_Report_Update();      // Update USB Keyboard report

        __disable_interrupt();      // Disable interrupts before checking 
                                    // if keyboard can sleep

        if (KBD430_Keyboard_CanSleep() == true)
        {
            // This example simply goes to LPM3 when sleeping
            // Optionally, gotoSleep  flag can be used to check if the host
            // is requesting Sleep Mode vs Wake-up
            __bis_SR_register(LPM3_bits + GIE);
            __no_operation();
        }
        
        __enable_interrupt();       // Re-enable interrupts 
    } // for (;;) 
}


//
// Private functions
//
/******************************************************************************
*
 * @brief   Clock initialization
 *  Initializes the clocks for the USB and rest of the system
 *
 * @return  none
 *****************************************************************************/
static void Init_Clock(void)
{
    if (CALBC1_8MHZ==0xFF)					// If calibration constant erased
    {											
      while(1);                               // do not load, trap CPU!!	
    }
    DCOCTL = 0;                               // Select lowest DCOx and MODx settings
    BCSCTL1 = CALBC1_8MHZ;
    DCOCTL = CALDCO_8MHZ;                     // Load 8MHz constants
    
    BCSCTL3 = LFXT1S_2;                         // Select VLO for ACLK
}

/******************************************************************************
*
 * @brief   GPIO initialization
 *  Initializes all the GPIOs
 *
 * @return  none
 *****************************************************************************/
static void Init_Ports(void)
{
    // KSI pins use external pull-ups, they could use internal pull-ups if desired
    // KSO pins must be initialized as output low
    // All unused pins as output low
    /* The definition of GPIOs for KBD430 board is as follows:
  P1:
    P1.0 KSI0           P1.1 KSI1               P1.2 KSI2               P1.3 KSI3
    P1.4 KSI4           P1.5 KSI5               P1.6 KSI6               P1.7 KSI7
  P2:
    P2.0 GPIO1_ACLK     P2.1 GPIO2_LED1         P2.2 GPIO3_LED2         P2.3 GPIO4_LED3
    P2.4 KSO0           P2.5 KSO1               P2.6 KSO2               P2.7 KSO3
  P3:
    P3.0 KSO4           P3.1 I2C_SDA            P3.2 I2C_SCL            P3.3 SPI_CLK
    P3.4 UART_TX        P3.5 UART_RX            P3.6 KSO5               P3.7 KSO6
  P4:
    P4.0 KSO7           P4.1 KSO8               P4.2 KSO9               P4.3 KSO10
    P4.4 KSO11          P4.5 KSO12              P4.6 KSO13              P4.7 KSO14
  */
    
    // Set GPIO functionality 
    P2SEL = 0x00;   // LFXT pins used as GPIOs

    // KSI pins as input (with external pull-up)
    P1DIR &= ~KSI_ALL;
    
    // KSO pins as output low
    P2OUT &= ~KSO_P2_ALL;
    P2DIR |= KSO_P2_ALL;
    P3OUT &= ~KSO_P3_ALL;
    P3DIR |= KSO_P3_ALL;
    P4OUT &= ~KSO_P4_ALL;
    P4DIR |= KSO_P4_ALL;
    
    // LED pins as output low by default
    P2OUT &= ~(LEDNUM_2_1|LEDCAPS_2_2|LEDSCROLL_2_3);
    P2DIR |= (LEDNUM_2_1|LEDCAPS_2_2|LEDSCROLL_2_3);
    
    // I2C pins as input (with external pull-up). 
    // They will be initialized for I2C functionality by hidi2c driver
    P3DIR &= ~(BIT1|BIT2);
    
    // GPIO_ACLK used to interrupt host. Set as input
    P2DIR &= ~BIT0;
 
}
