/** 
 *  @file   KBD430_customHID.c
 *  @brief  HID Custom interface protocol
 */
/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

//
// Include files
//
#include "CustomHID.h"
#include "KBD430_public_defs.h"
#include "KBD430_hardware.h"

#include "USB_config/descriptors.h"
#include "USB_API/USB_Common/usb.h"        // USB-specific functions
#include "usbConstructs.h"
#include <string.h>

#ifdef _HID_
    #include "USB_API/USB_HID_API/UsbHid.h"
#endif

#ifdef USE_CUSTOM_HID

//
// Local function prototypes
//
static void CustomHID_Parse(uint8_t * buffer);

//
// Local variables
//
/*! Local flag updated when there's new data on custom HID interface */
volatile bool bDataReceived_custom;
/*! Receive buffer */
__no_init static uint8_t RxBuff[CUSTOM_HID_RX_SIZE];

//
// Public functions
//
/******************************************************************************
 * @brief   Initializes the Custom HID protocol interface
 *
 * @return  none
 *****************************************************************************/
void KBD430_CustomHID_Init( void )
{
    uint8_t i;
    // Initialize variables and buffer
    bDataReceived_custom = false;
    for (i=0; i < CUSTOM_HID_RX_SIZE; i++)
    {
        RxBuff[i] = 0x00;
    }
}



/******************************************************************************
 * @brief  Polls the Custom HID interface
 *  Checks if host sent data to Custom interface, if so, parses and processes it
 *
 * @return  none
 *****************************************************************************/
void KBD430_CustomHID_Task( void )
{
    uint16_t count;

    // If data was received by the Event handler, then process it
    if (bDataReceived_custom == true)
    {
        // If data was received in "custom" interface, process it
        count = hidReceiveDataInBuffer((uint8_t *)RxBuff,
                                       CUSTOM_HID_RX_SIZE,
                                       HID_Custom);
        
        // Any packet bigger than 0 will be counted as valid
        if (count >= 1)
        {
            CustomHID_Parse(&RxBuff[0]);
        }
        bDataReceived_custom = false; // Clear the flag
    }

    return;
}


/******************************************************************************
 * @brief  Processes the data from custom HID interface
 *   This function is only included as an example and can be used as a base
 *   for different applications
 *
 * @param buffer pointer to the received data
 *
 * @return  none
 *****************************************************************************/
static void CustomHID_Parse(uint8_t *buffer)
{
    char outString[CUSTOM_HID_TX_SIZE] = "";     // Holds the outgoing string

    switch (buffer[0])
    {
        case '1':
            // toggle LED1
            LED_PORT_W ^= LEDNUM_1_1;
            // Prepare the outgoing string
            if (LED_PORT_W & LEDNUM_1_1)
                strcpy(outString,"NUM LED is ON\r\n");
            else
                strcpy(outString,"NUM LED is OFF\r\n");
        break;
        case '2':
            // toggle LED2
            LED_PORT_W ^= LEDCAPS_1_6;
            // Prepare the outgoing string
            if (LED_PORT_W & LEDCAPS_1_6)
                strcpy(outString,"CAPS LED is ON\r\n");
            else
                strcpy(outString,"CAPS LED is OFF\r\n");
        break;
        case '3':
            // toggle LED3
            LED_PORT_W ^= LEDSCROLL_1_7;
            // Prepare the LEDSCROLL_1_7 string
            if (LED_PORT_W & LEDSCROLL_1_7)
                strcpy(outString,"SCROLL LED is ON\r\n");
            else
                strcpy(outString,"SCROLL LED is OFF\r\n");
        break;
        default:
            // Invalid command
            strcpy(outString,"Invalid command\r\n\r\n");
        break;

    }
        // Wait until data is sent or error
         while (hidSendDataInBackground((uint8_t*)outString, strlen(outString), HID_Custom, 0) == 1)
             ;

}

#endif  //USE_CUSTOM_HID

