/** 
 *  @file   KBD430_main.c
 *  @brief  USB Keyboard main file
 */
/* --COPYRIGHT--,BSD
 * Copyright (c) 2014, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/

//
// Include files
//
#include "KBD430_Keyboard_public.h"
#include "CustomHID.h"
#include "KBD430_hardware.h"

#include "driverlib.h"
#include "USB_API/USB_Common/device.h"
#include "USB_API/USB_Common/types.h"      
#include "USB_config/descriptors.h"
#include "USB_API/USB_Common/usb.h"        
#include "usbConstructs.h"
#include "USB_API/USB_HID_API/UsbHid.h"

//
// Local function prototypes
//
static void Init_Clock(void);
static void Init_Ports(void);


//
// Global variables
//
/*! Flag set when data is received from custom interface */
volatile bool bDataReceived_kbd;  

//
// Public functions
//
/******************************************************************************
 * @brief   Main USB Keyboard function
 *  Initializes the MCU and then jumps to a main loop routine which:
 *    - Checks for pressed keys
 *    - Reports keyboard events to HID-keyboard interface
 *    - Polls HID-Custom interface and performs command as requested
 *
 * @return  none
 *****************************************************************************/
void main( void )
{
    bool   FirstTime;  // Indicates when keyboard is first connected to USB 
    uint8_t RxBuffKbd[MAX_PACKET_SIZE]; // Buffer to store received data

    // Stop watchdog timer to prevent time out reset
    WDT_A_hold(WDT_A_BASE);

    // Setup clock and core voltage
    PMM_setVCore(PMMCOREV_2);           // PMMCore = 2
    Init_Clock();                       // Init clocks
    Init_Ports();                       // Init GPIOs
    
#ifdef ENABLE_SW_BSL
    // Check if BSL is requested
    // This routine checks if 'MSP' is pressed after reset. 
    // If so, it calls BSL mode
    SET_KSO_POLL();                         // Set KSO as idle
    __delay_cycles(100000);                 // Wait for pull-ups
    KBD430_DKS_AssertKSO(8);                // Assert KSO for 'M'
    if (KBD430_DKS_IsKSIPressed(7) == true) // Check if KSI for 'M' is pressed
    {
        KBD430_DKS_DeAssertKSO(8);          // Deassert KSO for 'M'
        KBD430_DKS_AssertKSO(14);            // Assert KSO for 'S'
        if (KBD430_DKS_IsKSIPressed(3) == true) // Check if KSI for 'S' is pressed
        {
            KBD430_DKS_DeAssertKSO(14);         // Deassert KSO for 'S'
            KBD430_DKS_AssertKSO(4);           // Assert KSO for 'P'
            if (KBD430_DKS_IsKSIPressed(1) == true) // Check if KSI for 'P' is pressed
            {
                // Toggle LEDs to signal entry to BSL mode
                LED_PORT_W &= ~(LEDNUM_1_1|LEDCAPS_1_6|LEDSCROLL_1_7);
                __delay_cycles(2000000);
                LED_PORT_W |= (LEDNUM_1_1|LEDCAPS_1_6|LEDSCROLL_1_7);
                __delay_cycles(2000000);
                LED_PORT_W &= ~(LEDNUM_1_1|LEDCAPS_1_6|LEDSCROLL_1_7);
                __disable_interrupt();
                ((void (*)())0x1000)();         // Jump to BSL
            }
        }
    }
    SET_KSO_INTERRUPT();                        // Set KSO for interrupt
#endif

    // Configure SVSM
    // SVSL = OFF
    // SVML = OFF
    // SVSH = Full Performance , causes Reset, stays on in LPM3
    // SVSL = OFF
    PMM_disableSvsL();  
    PMM_disableSvmL();  
    PMM_disableSvmH();  
    PMM_enableSvsH();   
    PMM_enableSvsHReset();
    PMM_SvsHEnabledInLPMFullPerf();   
                                          
    // Wait until high side, low side settled
    while ( 0 == PMM_getInterruptStatus(SVSMLDLYIFG + SVSMHDLYIFG ) ) 
        ;
    PMM_clearPMMIFGS();

    // Init USB
    USB_init();     
    // Enable all USB-level events
    USB_setEnabledEvents(kUSB_VbusOnEvent+kUSB_VbusOffEvent+kUSB_dataReceivedEvent+
                        kUSB_receiveCompletedEvent+kUSB_UsbSuspendEvent+kUSB_UsbResumeEvent);

#ifdef USE_CUSTOM_HID
    KBD430_CustomHID_Init();            // Initialize Custom HID interface protocol
#endif
    KBD430_Keyboard_Init();             // Initialize Digital Keyscan
    
    // Clear global HID receive flags
    bDataReceived_kbd = false;

    // Set to signal first loop
    FirstTime = true;

    __enable_interrupt();   

    // In case USB is already attached (meaning no VBUS event will
    // occur), manually start the connection
    if (USB_connectionInfo() & kUSB_vbusPresent)
    {
        USB_handleVbusOnEvent();
    }


    for (;;)
    {
        switch(USB_connectionState())
        {
            case (ST_USB_DISCONNECTED | ST_NOENUM_SUSPENDED):
                // Enter LPM3 w/interrupt.  Nothing for us to do while disconnected.
                FirstTime = true;
                KBD430_DKS_Disable();
                __bis_SR_register(LPM3_bits + GIE);     
            break;

            case ST_USB_CONNECTED_NO_ENUM:
                FirstTime = true;
                KBD430_DKS_Disable();
            break;

            case ST_ENUM_ACTIVE:
                // This is optional, but it shows when the keyboard is ready to go
                if (FirstTime == true)
                {
                    // Toggle LEDs just to signal that keyboard is ready
                    LED_PORT_W &= ~(LEDNUM_1_1|LEDCAPS_1_6|LEDSCROLL_1_7);
                    __delay_cycles(2000000);
                    LED_PORT_W |= (LEDNUM_1_1);
                    __delay_cycles(2000000);
                    LED_PORT_W |= (LEDNUM_1_1|LEDCAPS_1_6);
                    __delay_cycles(2000000);
                    LED_PORT_W |= (LEDNUM_1_1|LEDCAPS_1_6|LEDSCROLL_1_7);
                    __delay_cycles(2000000);
                    LED_PORT_W &= ~(LEDNUM_1_1|LEDCAPS_1_6|LEDSCROLL_1_7);

                    // Re-initialize modules
                    KBD430_Keyboard_Init();  // Initialize keyboard
                    
                    // Don't re-initialize unless we are disconnected
                    FirstTime = false;        
                }

#ifdef USE_CUSTOM_HID                
                // Poll the USB Custom interface
                KBD430_CustomHID_Task();
#endif

                // This flag is true if the Standard keyboard interface gets data
                if (bDataReceived_kbd == true)
                {
                    // If there's new data, get the report
                    if (USBHID_receiveReport((uint8_t *)RxBuffKbd, HID_Keyboard) >= 1)
                    {
                        // Check for LED usage data and turn on/off NUM, CAPS & SCROLL LEDs
                        if (RxBuffKbd[0] & 0x01)
                        {
                            LED_PORT_W |= (LEDNUM_1_1);
                        }
                        else
                        {
                            LED_PORT_W &= ~(LEDNUM_1_1);
                        }

                        if (RxBuffKbd[0] & 0x02)
                        {
                            LED_PORT_W |= (LEDCAPS_1_6);
                        }
                        else
                        {
                            LED_PORT_W &= ~(LEDCAPS_1_6);
                        }

                        if (RxBuffKbd[0] & 0x04)
                        {
                            LED_PORT_W |= (LEDSCROLL_1_7);
                        }
                        else
                        {
                            LED_PORT_W &= ~(LEDSCROLL_1_7);
                        }
                    }
                    // Clear flag in order to process next time
                    bDataReceived_kbd = false;
                }

                KBD430_Keyboard_Task();     // Task checking keyboard
                
                KBD430_Report_Update();     // Update USB Keyboard report
                
                __disable_interrupt();      // Disable interrupts before checking 
                                            //  if module can sleep
                // Enter LPM0 w/interrupt if nothing is pending
                if (KBD430_Keyboard_CanSleep() == true)
                {
                    __bis_SR_register(LPM0_bits + GIE);     
                    __no_operation();
                }
                
                __enable_interrupt();       // Re-enable interrupts 
            break;

            case ST_ENUM_SUSPENDED:
                // suspended.  
                FirstTime = true;   
                // Enter LPM3 w/interrupt
              __bis_SR_register(LPM3_bits + GIE);   
              
              USB_forceRemoteWakeup();  // Force wake-up if waking-up from keyboard
            break;
            
            case ST_ENUM_IN_PROGRESS:
            case ST_ERROR:
                FirstTime = true;
                KBD430_DKS_Disable();
            break;

            default:;
        } // switch(USB_connectionState())
    } // for (;;) 
}


//
// Private functions
//
/******************************************************************************
*
 * @brief   Clock initialization
 *  Initializes the clocks for the USB and rest of the system
 *
 * @return  none
 *****************************************************************************/
static void Init_Clock(void)
{
    // Use REFO (32.768Khz) as source for FLL 
	UCS_clockSignalInit(
	   UCS_FLLREF,
	   UCS_REFOCLK_SELECT,
	   UCS_CLOCK_DIVIDER_1);

    // Use REFO (32.768Khz) as source for ACLK
	UCS_clockSignalInit(
	   UCS_ACLK,
	   UCS_REFOCLK_SELECT,
	   UCS_CLOCK_DIVIDER_1);

    // Set MCLK=SMCLK at the specified MCLK_FREQ frequency
    UCS_initFLLSettle(
        MCLK_FREQ/1000,
        MCLK_FREQ/32768);

}

/******************************************************************************
*
 * @brief   GPIO initialization
 *  Initializes all the GPIOs
 *
 * @return  none
 *****************************************************************************/
static void Init_Ports(void)
{
    // KSI pins use external pull-ups, they could use internal pull-ups if desired
    // KSO pins must be initialized as output low
    // All unused pins as output low
    /* The definition of GPIOs for KBD430 board is as follows:
  P1:
    P1.0 GPIO1_ACLK     P1.1 GPIO2_LED1         P1.2 TCAP1              P1.3 TCAP2
    P1.4 PWM1           P1.5 PWM2               P1.6 GPIO3_LED2         P1.7 GPIO4_LED3
  P2:
    P2.0 KSI0           P2.1 KSI1               P2.2 KSI2               P2.3 KSI3
    P2.4 KSI4           P2.5 KSI5               P2.6 KSI6               P2.7 KSI7
  P3:
    P3.0 SPI_MOSI       P3.1 SPI_MISO           P3.2 SPI_CLK            P3.3 UART_TX
    P3.4 UART_RX        P3.5 GPIO5              P3.6 GPIO6              P3.7 GPIO7
  P4:
    P4.0 GPIO8          P4.1 I2C_SDA            P4.2 I2C_SCL            P4.3 GPIO9
    P4.4 GPIO10         P4.5 GPIO11             P4.6 GPIO12             P4.7 KSO0
  P5:
    P5.0 AIN7           P5.1 AIN8               P5.2 XT2IN              P5.3 XT2OUT
    P5.4 KSO1           P5.5 KSO2               P5.6 KSO3               P5.7 KSO4
  P6:
    P6.0 AIN1           P6.1 AIN2               P6.2 AIN3               P6.3 AIN4
    P6.4 AIN5           P6.5 AIN6               P6.6 KSO5               P6.7 KSO6
  P7:
    P7.0 KSO7           P7.1 KSO8               P7.2 KSO9               P7.3 KSO10
    P7.4 PWM3           P7.5 PWM4               P7.6 PWM5               P7.7 KSO11
  P8:
    P8.0 KSO12          P8.1 KSO13              P8.2 KSO14
  PJ:
    PJ.0 GPIO13         PJ.1 N/C                PJ.2 N/C                PJ.3 N/C
  */

    // KSI pins as input
    GPIO_setAsInputPin(GPIO_PORT_P2, KSI_ALL);
    
    // KSO pins as output low
    GPIO_setOutputLowOnPin(GPIO_PORT_P4, KSO_P4_ALL);
    GPIO_setAsOutputPin(GPIO_PORT_P4, KSO_P4_ALL);
    GPIO_setOutputLowOnPin(GPIO_PORT_P5, KSO_P5_ALL);
    GPIO_setAsOutputPin(GPIO_PORT_P5, KSO_P5_ALL);
    GPIO_setOutputLowOnPin(GPIO_PORT_P6, KSO_P6_ALL);
    GPIO_setAsOutputPin(GPIO_PORT_P6, KSO_P6_ALL);
    GPIO_setOutputLowOnPin(GPIO_PORT_P7, KSO_P7_ALL);
    GPIO_setAsOutputPin(GPIO_PORT_P7, KSO_P7_ALL);
    GPIO_setOutputLowOnPin(GPIO_PORT_P8, KSO_P8_ALL);
    GPIO_setAsOutputPin(GPIO_PORT_P8, KSO_P8_ALL);

    // LED pins as output low by default
    GPIO_setOutputLowOnPin(GPIO_PORT_P1, LEDNUM_1_1|LEDCAPS_1_6|LEDSCROLL_1_7);
    GPIO_setAsOutputPin(GPIO_PORT_P1, LEDNUM_1_1|LEDCAPS_1_6|LEDSCROLL_1_7);
    
    // Unused pins as output low
    GPIO_setOutputLowOnPin(GPIO_PORT_P1, BIT0|BIT2|BIT3|BIT4|BIT5);
    GPIO_setAsOutputPin(GPIO_PORT_P1, BIT0|BIT2|BIT3|BIT4|BIT5);
    GPIO_setOutputLowOnPin(GPIO_PORT_P3, BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);
    GPIO_setAsOutputPin(GPIO_PORT_P3, BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6|BIT7);
    GPIO_setOutputLowOnPin(GPIO_PORT_P4, BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6);
    GPIO_setAsOutputPin(GPIO_PORT_P4, BIT0|BIT1|BIT2|BIT3|BIT4|BIT5|BIT6);
    GPIO_setOutputLowOnPin(GPIO_PORT_P5, BIT0|BIT1|BIT2|BIT3);
    GPIO_setAsOutputPin(GPIO_PORT_P5, BIT0|BIT1|BIT2|BIT3);
    GPIO_setOutputLowOnPin(GPIO_PORT_P6, BIT0|BIT1|BIT2|BIT3|BIT4|BIT5);
    GPIO_setAsOutputPin(GPIO_PORT_P6, BIT0|BIT1|BIT2|BIT3|BIT4|BIT5);
    GPIO_setOutputLowOnPin(GPIO_PORT_P7, BIT4|BIT5|BIT6);
    GPIO_setAsOutputPin(GPIO_PORT_P7, BIT4|BIT5|BIT6);
    GPIO_setOutputLowOnPin(GPIO_PORT_PJ, BIT0|BIT1|BIT2|BIT3);
    GPIO_setAsOutputPin(GPIO_PORT_PJ, BIT0|BIT1|BIT2|BIT3);
   
}

/******************************************************************************
 *
 * @brief   System Non-Maskable interrupt fault ISR
 *  Handles system interrupts
 *
 * @return  none
 *****************************************************************************/
#pragma vector = UNMI_VECTOR
__interrupt void UNMI_ISR(void)
{
    switch (__even_in_range(SYSUNIV, SYSUNIV_BUSIFG))
    {
        case SYSUNIV_NONE:
            __no_operation();
            break;
        case SYSUNIV_NMIIFG:
            __no_operation();
            break;
        case SYSUNIV_OFIFG:
            UCS_clearFaultFlag(UCS_XT2OFFG);
            UCS_clearFaultFlag(UCS_DCOFFG);
            SFR_clearInterrupt(SFR_OSCILLATOR_FAULT_INTERRUPT);
            break;
        case SYSUNIV_ACCVIFG:
            __no_operation();
            break;
        case SYSUNIV_BUSIFG:
            // If the CPU accesses USB memory while the USB module is
            // suspended, a "bus error" can occur.  This generates an NMI.  If
            // USB is automatically disconnecting in your software, set a
            // breakpoint here and see if execution hits it.  See the
            // Programmer's Guide for more information.
            SYSBERRIV = 0; //clear bus error flag
            USB_disable(); //Disable
    }
} 
