/* --COPYRIGHT--,BSD
 * Copyright (c) 2012, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *    its contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * --/COPYRIGHT--*/
/**
 * @file i2c_engine_usci.c
 * @brief I2C (USCI Implementation) for HIDI2C Specification
 * I2C engine for the HIDI2C implementation. 
 * @details This is the transport layer of the API. In this implementation, USCI 
 * is used for management of the I2C. This layer is abstracted out so another 
 * implementation of I2C can be used. 
 *
 * GPIO ports and specific interface settings are defined in the @link
 * hidi2c_settings.h @endlink file.
 *
 */
#include <msp430.h>

#include <hidi2c_settings.h>
#include <hidi2c_errno.h>
#include "hid_engine.h"

/* Interface Settings */
#include <i2c_engine_usci.h>
#include <i2c_engine_usci_init.h>

#include <stdlib.h>
#include <string.h>
#include <stdio.h>

/* Local Variables */
static volatile BOOL clearLPMBits;

/* External call to HID layer */
extern BOOL hidI2C_handleData(uint8_t interface, uint8_t* data, uint16_t reg);
extern void hidI2C_handleRegisterPointer(uint8_t interface,
        uint16_t currentRegisterIndex);
extern void hidI2C_setNextQueuedData(uint8_t inf);
extern BOOL hidI2C_haveMoreGPIOsPending(uint8_t inf);

#ifdef USCIB0
uint16_t usciB0BytesToXfer;
uint8_t* usciB0CurDataPointer;
uint16_t usciB0xferIndex;
uint8_t usciB0RegisterPointer[2] =
{ 0, 0 };
uint16_t usciB0CurrentRegister;
BOOL usciB0DeviceInitiated;
BOOL usciB0startedOnce;
#endif


/**
 * Initializes the I2C physical layer of the HIDI2C API. ifdefs are used to
 * minimize the memory footprint if the user is not using every interface.
 *
 * @return TRUE if initialization succeeded. FALSE otherwise.
 */
BOOL hidI2C_initializeI2C(BOOL idle)
{
    /* Setting up I2C */
    hidI2C_initializeUSCII2C(idle);

#ifdef USCIB0
    IE2 |= UCB0RXIE | UCB0TXIE;         // Enable RX/TX interrupts 
    UCB0I2CIE |= UCSTTIE;               // Enable Start interrupt
    usciB0BytesToXfer = 0;
    usciB0CurDataPointer = 0;
    usciB0startedOnce = FALSE;
    usciB0DeviceInitiated = FALSE;
    usciB0xferIndex = 0;
    usciB0CurrentRegister = 0;
#endif


    return TRUE;
}

/**
 * Internal API command to transfer data from the host and initialize interrupt.
 * Checking for conflicts concurrent accesses is done in the HID layer.
 *
 * @param interface Specific USCI module to send data over. Defined in the
 *  @link hidi2c_settings.h @endlink file.
 * @param data Pointer to data to "queue" for transfer.
 */
uint8_t hidI2C_setDataAndInitializeInterrupt(uint8_t interface, uint8_t* data,
        uint16_t length)
{
    switch (interface)
    {
#ifdef USCIB0
        case USCIB0:
            usciB0CurDataPointer = data;
            usciB0xferIndex = 0;
            usciB0BytesToXfer = length;
            break;
#endif

    }

    hidI2C_setGPIOPinActive(interface);

    return HIDI2C_SUCCESS;
}

/**
 * Function that sets the current data pointer without changing the register
 * value.
 */
uint8_t hidI2C_setDataWithoutInterrupt(uint8_t interface, uint8_t* data,
        uint16_t length)
{
    return hidI2C_setDataChangeRegWithoutInterrupt(interface, data, length, 0);
}

/**
 * Internal API command to transfer data from the host without interrupt.
 * @param interface Specific USCI module to send data over. Defined in the
 *  @link hidi2c_settings.h @endlink file.
 * @param data Pointer to data to "queue" for transfer.
 */
uint8_t hidI2C_setDataChangeRegWithoutInterrupt(uint8_t interface,
        uint8_t* data, uint16_t length, uint16_t registerIndex)
{
    switch (interface)
    {
#ifdef USCIB0
        case USCIB0:
            usciB0CurDataPointer = data;
            usciB0BytesToXfer = length;
            usciB0xferIndex = 0;

            if (registerIndex != 0)
                usciB0CurrentRegister = registerIndex;

            break;
#endif

    }

    return HIDI2C_SUCCESS;
}

/**
 * Returns true if the GPIO is asserted for the given interface.
 */
BOOL hidI2C_isGPIOAsserted(uint8_t interface)
{
    switch (interface)
    {
#ifdef USCIB0
        case USCIB0:
            return (USCIB0_GPIO_PDIR & USCIB0_GPIO_PIN);
#endif

    }

    return FALSE;
}

/**
 * Function that returns true if there is a transfer in progress on given
 * interface
 * @param interface Specific USCI module to send data over. Defined in the
 *  @link hidi2c_settings.h @endlink file.
 *
 * @return TRUE if a transfer is in progress, false otherwise.
 */
BOOL hidI2C_busyWithTransfer(uint8_t interface)
{
    switch (interface)
    {
#ifdef USCIB0
        case USCIB0:
            return (UCB0STAT & UCBUSY);
#endif

    }

    return FALSE;
}

/**
 * Disables START interrupts from happening on given interface.
 * @param interface Specific USCI module to send data over. Defined in the
 *  @link hidi2c_settings.h @endlink file.
 */
void hidI2C_disableInterrupts(uint8_t interface)
{
    switch (interface)
    {
#ifdef USCIB0
        case USCIB0:
            UCB0I2CIE &= ~UCSTTIE;
            break;
#endif
    }
}

/**
 * Allows START interrupts to happen on given interface
 * @param interface Specific USCI module to send data over. Defined in the
 *  @link hidi2c_settings.h @endlink file.
 */
void hidI2C_enableInterrupts(uint8_t interface)
{
    switch (interface)
    {
#ifdef USCIB0
        case USCIB0:
            UCB0I2CIE |= UCSTTIE;
            break;
#endif

    }
}

/**
 * Simple function to take two bytes and combine them into one word. This is
 * useful for HIDI2C because everything comes in from the host MSB first
 * @param msb Most significant byte of data 
 * @param lsb Least significant byte of data
 */
uint16_t bitSwap(uint8_t msb, uint8_t lsb)
{
    uint16_t res = 0;
    uint16_t tmpBuffer = 0;

    res = lsb;
    tmpBuffer = msb;
    tmpBuffer = tmpBuffer << 8;
    res = res | tmpBuffer;

    return res;
}

/*
 * USCI_B0 RXVECTOR handles the STT/STP flags
 */
#ifdef USCIB0
#ifdef __GNUC__
__attribute__((interrupt(USCIAB0RX_VECTOR)))
#else
#pragma vector=USCIAB0RX_VECTOR
__interrupt
#endif
void USCI_B0_FLAGS_ISR(void)
{
    if (UCB0STAT & UCSTTIFG)        // Start Condition flag
    {
        usciB0xferIndex = 0;

        /* Disabling start interrupts for the time being */
        hidI2C_disableInterrupts(USCIB0);

        /* First Start Condition */
        if (!usciB0startedOnce)
        {
            usciB0BytesToXfer = 0;

            /* Initial RXing */
            if (!(UCB0CTL1 & UCTR))
            {
                usciB0BytesToXfer = 2;
                usciB0CurDataPointer = usciB0RegisterPointer;
            }
            /* Initial TXing */
            else
            {
                hidI2C_setNextQueuedData(USCIB0);
                usciB0DeviceInitiated = TRUE;
                usciB0xferIndex = 0;
                UCB0I2CIE |= (UCSTPIE);
            }

            usciB0startedOnce = TRUE;
        }
        UCB0STAT &= ~UCSTTIFG;
    }
    else if (UCB0STAT & UCSTPIFG)           // Stop Condition flag
    {
        if (!hidI2C_haveMoreGPIOsPending(USCIB0))
        {
          hidI2C_setGPIOPinIdle(USCIB0);
        }
        UCB0I2CIE &= ~(UCSTPIE);
        UCB0STAT &= ~UCSTPIFG;
    }
}


/*
 * USCI_B0 TXVECTOR handles the RX/TX flags
 */
#ifdef __GNUC__
__attribute__((interrupt(USCIAB0TX_VECTOR)))
#else
#pragma vector=USCIAB0TX_VECTOR
__interrupt
#endif
void USCI_B0_RXTX_ISR(void)
{
    if (IFG2 & UCB0RXIFG)               // RX Flag
    {
#ifdef USCI30WORKAROUND
        /* ERATTA USCI30 */
        uint8_t bLastStatus, bCount;
        
        bLastStatus = UCSCLLOW & UCB0STAT;
        bCount = 0;
    
        while (bCount < 12)
        {
            if(bLastStatus != (UCSCLLOW & UCB0STAT))
            {
                bLastStatus = UCSCLLOW & UCB0STAT;
                bCount = 0;
            }
            else
            {
                //(delay is 1 quarter of bit clock)
                __delay_cycles(HIDI2CUSCI30DELAY);
                bLastStatus = UCSCLLOW & UCB0STAT;
                bCount++;
            }
        }
#endif

        usciB0CurDataPointer[usciB0xferIndex] = UCB0RXBUF;
        usciB0xferIndex++;

        /* This is the last byte */
        if (usciB0xferIndex == usciB0BytesToXfer)
        {
            /* First two bytes */
            if (usciB0CurDataPointer == usciB0RegisterPointer)
            {
                usciB0CurrentRegister = bitSwap(usciB0RegisterPointer[1],
                        usciB0RegisterPointer[0]);
                hidI2C_handleRegisterPointer(USCIB0, usciB0CurrentRegister);
            }
            /* Something else */
            else
            {
                usciB0BytesToXfer = 0;

                clearLPMBits = hidI2C_handleData(USCIB0,
                        usciB0CurDataPointer, usciB0CurrentRegister);

                if (usciB0BytesToXfer == 0)
                {
                    usciB0startedOnce = FALSE;
                    hidI2C_enableInterrupts(USCIB0);
                }
            }
        }
    }
    else if (IFG2 & UCB0TXIFG)                  // TX Flag
    {
        UCB0TXBUF = usciB0CurDataPointer[usciB0xferIndex];
        usciB0xferIndex++;

        /* Last Byte to TX */
        if (usciB0xferIndex == usciB0BytesToXfer)
        {
            if (usciB0DeviceInitiated)
            {
                usciB0DeviceInitiated = FALSE;
            }

            usciB0startedOnce = FALSE;
            hidI2C_enableInterrupts(USCIB0);
        }

    }
    
    /* Clearing LPM bits if we need to */
    if (clearLPMBits)
    {
        clearLPMBits = FALSE;
        __bic_SR_register_on_exit(LPM4_bits);
        return;
    }

}
#endif

